import operator
from functools import reduce

from django.db.models import Q

from .core import Autocomplete


class ModelAutocomplete(Autocomplete):
    model = None
    search_attrs = []

    @classmethod
    def get_search_attrs(cls):
        if not cls.search_attrs:
            raise ValueError("ModelAutocomplete must have search_attrs")
        return cls.search_attrs

    @classmethod
    def get_model(cls):
        if not cls.model:
            raise ValueError("ModelAutocomplete must have a model")

        return cls.model

    @classmethod
    def get_queryset(cls):
        return cls.get_model().objects.all()

    @classmethod
    def get_label_for_record(cls, record):
        return str(record)

    @classmethod
    def get_query_filtered_queryset(cls, search, context):
        base_qs = cls.get_queryset()
        conditions = [
            Q(**{f"{attr}__icontains": search}) for attr in cls.get_search_attrs()
        ]
        condition_filter = reduce(operator.or_, conditions)
        queryset = base_qs.filter(condition_filter)
        return queryset

    @classmethod
    def search_items(cls, search, context):
        filtered_queryset = cls.get_query_filtered_queryset(search, context)

        items = QuerysetMappedIterable(
            queryset=filtered_queryset, label_for_record=cls.get_label_for_record
        )
        return items

    @classmethod
    def get_items_from_keys(cls, keys, context):
        queryset = cls.get_queryset()
        results = queryset.filter(id__in=keys)

        return [
            {"key": record.id, "label": cls.get_label_for_record(record)}
            for record in results
        ]


class QuerysetMappedIterable:
    """
    We want to return an iterable of dicts rather than ORM records

    Using a list is inefficient for large datasets
    But using something like a generator/map doesn't allow for slicing or len()

    This class wraps a queryset's slice/len methods
    """

    def __init__(self, queryset, label_for_record):
        self.queryset = queryset
        self.label_for_record = label_for_record

    def __iter__(self, *args, **kwargs):
        return (self.map_record(r) for r in self.queryset)

    def map_record(self, record):
        return {"key": record.id, "label": self.label_for_record(record)}

    def __getitem__(self, key):
        # Handle both single index and slice objects
        if isinstance(key, int):
            records = [self.queryset[key]]
        elif isinstance(key, slice):
            records = self.queryset[key.start : key.stop : key.step]
        else:
            raise TypeError("Invalid argument type")

        mapped = [self.map_record(r) for r in records]

        if isinstance(key, int):
            return mapped[0]

        return mapped

    def __len__(self):
        # Return the length of the sequence
        return self.queryset.count()



"""
This file enables the component to be used like other Django widgets
"""

from django.forms import Widget

from .core import AC_CLASS_CONFIGURABLE_VALUES, Autocomplete


class AutocompleteWidget(Widget):
    template_name = "autocomplete/component.html"

    configurable_values = [
        "indicator",
        "multiselect",
        "label",
        "component_prefix",
        # the below are also configurable from the AC class
        "placeholder",
    ]

    def __init__(self, ac_class, attrs=None, options=None):
        self.ac_class = ac_class
        super().__init__(attrs)

        if not options:
            options = {}

        self.config = {}
        for k, v in options.items():
            if k in self.configurable_values:
                self.config[k] = v
            else:
                raise ValueError(f"Invalid option {k}")

    def value_from_datadict(self, data, files, name):
        if self.is_multi:
            try:
                # classic POSTs go though django's QueryDict structure
                # which has a getlist method
                value = data.getlist(name)
            except AttributeError:
                # some clients just provide lists in JSON
                value = data.get(name)

        else:
            value = data.get(name)

        return value

    def value_omitted_from_data(self, data, files, name):
        # An unselected <select multiple> doesn't appear in POST data, so it's
        # never known if the value is actually omitted.
        return []

    def get_component_id(self, field_name):
        prefix = self.get_configurable_value("component_prefix")

        return prefix + field_name

    def get_configurable_value(self, key):
        if key in self.config:
            return self.config.get(key)

        if key in AC_CLASS_CONFIGURABLE_VALUES and hasattr(self.ac_class, key):
            return getattr(self.ac_class, key)

        return None

    @property
    def is_multi(self):
        return self.get_configurable_value("multiselect")

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        proper_attrs = self.build_attrs(self.attrs, attrs)

        if value is None:
            selected_options = []
        else:
            if self.is_multi:
                selected_options = self.ac_class.get_items_from_keys(value, None)
            else:
                selected_options = self.ac_class.get_items_from_keys([value], None)

        context["ac_class"] = self.ac_class
        context["field_name"] = name
        context["id"] = attrs.get("id", self.attrs.get("id", None))
        context["route_name"] = self.ac_class.route_name

        context["disabled"] = proper_attrs.get("disabled", False)
        context["required"] = proper_attrs.get("required", False)

        context["indicator"] = self.get_configurable_value("indicator")
        context["multiselect"] = self.is_multi

        context["label"] = self.get_configurable_value("label")
        context["placeholder"] = self.get_configurable_value("placeholder")
        # context["values"] = list(self.a_c.item_values(self.a_c, selected_options))
        context["values"] = [x["key"] for x in selected_options]
        context["selected_items"] = selected_options
        context["component_prefix"] = self.get_configurable_value("component_prefix")
        context["component_id"] = self.get_component_id(name)

        return context



from django import forms
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.template import loader

from autocomplete import Autocomplete, AutocompleteWidget, ModelAutocomplete, register

from .models import Person, Team


@register
class PersonAutocomplete(Autocomplete):
    @classmethod
    def search_items(cls, search, context):
        qs = Person.objects.filter(name__icontains=search)

        return [{"key": person.id, "label": person.name} for person in qs]

    @classmethod
    def get_items_from_keys(cls, keys, context):
        qs = Person.objects.filter(id__in=keys)
        return [{"key": person.id, "label": person.name} for person in qs]


class TeamForm(forms.ModelForm):
    # this form isn't meant to work for saving, we're using different "names"
    class Meta:
        model = Team
        fields = ["team_lead", "members"]
        widgets = {
            "team_lead": AutocompleteWidget(
                ac_class=PersonAutocomplete,
            ),
            "members": AutocompleteWidget(
                ac_class=PersonAutocomplete,
                options={"multiselect": True},
            ),
        }


def edit_team(request, team_id=None):
    team = Team.objects.get(id=team_id)

    form = TeamForm(instance=team, data=request.POST or None)

    if request.POST and form.is_valid():
        form.save()
        return HttpResponseRedirect(request.path)

    return render(request, "edit_team.html", {"form": form})


@register
class CustomPersonAutocomplete(PersonAutocomplete):
    no_result_text = "Keine resultate"
    narrow_search_text = "NARROW IT DOWN"
    max_results = 1
    placeholder = "Select team lead!"


@register
class CustomPersonAutocomplete2(PersonAutocomplete):
    """
    this AC is meant to be used in a form with a 'team_lead' field

    the extra-hx vals will submit an additional field when searching

    this extra param will be used by the search method
    to filter out the team_lead from the members
    """

    @classmethod
    def get_extra_text_input_hx_vals(cls):
        # single quotes not allowed here, backticks used as 2nd level quotes
        return {
            "related_team_lead": 'document.querySelector(`[name="team_lead"]`)?.value || "" ',
            # "literal": "foo", # note that this causes 'ReferenceError: foo is not defined'
            "literal": '"foo"',  # wrapping in double quotes works
        }

    @classmethod
    def search_items(cls, search, context):
        qs = Person.objects.filter(name__icontains=search)

        related_team_lead = context.client_kwargs.get("related_team_lead", None)
        if related_team_lead:
            qs = qs.exclude(id=related_team_lead)

        return [{"key": person.id, "label": person.name} for person in qs]


class TeamForm2(forms.ModelForm):
    # this form isn't meant to work for saving, we're using different "names"
    class Meta:
        model = Team
        fields = ["team_lead", "members"]
        # fields = ["team_lead"]
        widgets = {
            "team_lead": AutocompleteWidget(
                ac_class=CustomPersonAutocomplete,
                options={
                    "component_prefix": "team_lead_prefix",
                    # "placeholder": "Select team lead",
                },
                attrs={
                    "required": False,
                },
            ),
            "members": AutocompleteWidget(
                ac_class=CustomPersonAutocomplete2,
                options={"multiselect": True},
            ),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["team_lead"].required = False
        # self.fields["members"].required = False
        # self.fields["members"].disabled = True
        # self.fields["team_lead"].disabled = True


def example_with_prefix(request, team_id=None):
    team = Team.objects.get(id=team_id)

    form = TeamForm2(instance=team, data=request.POST or None)

    if request.POST and form.is_valid():
        form.save()
        return HttpResponseRedirect(request.path)

    return render(request, "edit_team.html", {"form": form})


@register
class CustomPersonAutocomplete3(ModelAutocomplete):
    model = Person
    search_attrs = ["name"]


class AnotherPersonForm(forms.ModelForm):
    class Meta:
        model = Team
        fields = ["team_lead"]
        widgets = {
            "team_lead": AutocompleteWidget(
                ac_class=CustomPersonAutocomplete3,
            )
        }


def example_with_model(request, team_id=None):
    team = Team.objects.get(id=team_id)

    form = AnotherPersonForm(instance=team, data=request.POST or None)

    if request.POST and form.is_valid():
        form.save()
        return HttpResponseRedirect(request.path)

    return render(request, "edit_team.html", {"form": form})



import factory
from django.db import models


class Person(models.Model):
    name = models.CharField(max_length=60)

    def __str__(self):
        return self.name


class Team(models.Model):
    name = models.CharField(max_length=60)
    team_lead = models.ForeignKey(
        Person, null=True, on_delete=models.SET_NULL, related_name="lead_teams"
    )
    members = models.ManyToManyField(Person, related_name="teams")


class PersonFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Person

    name = factory.Faker("name")


class TeamFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Team

    name = factory.Faker("name")



import pytest
from django import forms
from django.template import Context, Template, loader
from django.urls import reverse

from autocomplete import Autocomplete, AutocompleteWidget, ModelAutocomplete, register
from autocomplete.core import ContextArg
from sample_app.models import Person, PersonFactory, Team, TeamFactory

from .utils_for_test import soup_from_str


class PersonModelAC(ModelAutocomplete):
    model = Person
    search_attrs = ["name"]


def test_model_ac_search():
    p1 = PersonFactory(name="John1")
    p2 = PersonFactory(name="John2")
    p3 = PersonFactory(name="John3")
    p4 = PersonFactory(name="Jones")

    results = PersonModelAC.search_items("Joh", ContextArg(None, None))

    assert len(results) == 3
    assert list(results) == [
        {"label": "John1", "key": p1.id},
        {"label": "John2", "key": p2.id},
        {"label": "John3", "key": p3.id},
    ]

    assert PersonModelAC.get_items_from_keys([p1.id, p2.id], {}) == [
        {"label": "John1", "key": p1.id},
        {"label": "John2", "key": p2.id},
    ]


def test_model_ac_search_max_results(django_assert_max_num_queries):
    class PersonModelAC(ModelAutocomplete):
        model = Person
        search_attrs = ["name"]
        max_results = 2

    p1 = PersonFactory(name="John1")
    p2 = PersonFactory(name="John2")
    p3 = PersonFactory(name="John3")
    p4 = PersonFactory(name="Jones")

    with django_assert_max_num_queries(1):
        results = PersonModelAC.search_items("Joh", ContextArg(None, None))
    # should still contain 3 results, the view is responsible for truncating
    assert len(results) == 3


def test_num_queries(django_assert_max_num_queries):
    """
    regression: the queryset iterable wrapper was repeatedly evaluating the queryset
    """

    p1 = PersonFactory(name="John1")
    p2 = PersonFactory(name="John2")
    p3 = PersonFactory(name="John3")
    p4 = PersonFactory(name="Jones")

    with django_assert_max_num_queries(1):
        results = PersonModelAC.search_items("Joh", ContextArg(None, None))
        mapped_results = PersonModelAC.map_search_results(results, [])



"""
Allow HTMXAutoComplete to be imported from the module directly
"""

from .core import Autocomplete, register
from .shortcuts import ModelAutocomplete
from .views import urls
from .widgets import AutocompleteWidget



import json

from django.http import QueryDict
from django.urls import reverse

from autocomplete.core import Autocomplete, register
from sample_app.models import Person, PersonFactory, Team, TeamFactory

from .utils_for_test import get_soup


class PersonAC3(Autocomplete):

    @classmethod
    def search_items(cls, search, context):
        qs = Person.objects.filter(name__icontains=search)

        return [{"key": person.id, "label": person.name} for person in qs]

    @classmethod
    def get_items_from_keys(cls, keys, context):
        qs = Person.objects.filter(id__in=keys)
        return [{"key": person.id, "label": person.name} for person in qs]


register(PersonAC3)


def test_toggle_response_select_from_empty_non_multi(client):
    """
    first, try adding selectin a person when none is selected
    """
    people = PersonFactory.create_batch(5)
    to_add = PersonFactory()

    base_url = reverse("autocomplete:toggle", kwargs={"ac_name": "PersonAC3"})
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "item": to_add.id,
        }
    )

    response = client.get(f"{base_url}?{qs_dict.urlencode()}")
    assert response.status_code == 200

    soup = get_soup(response)

    # 1. The element we are toggling (not sure why this is even included)
    toggled_option = soup.select("a[role='option']")
    assert len(toggled_option) == 1

    # 2. The hidden inputs that actually hold the form values
    hidden_inputs_container = soup.select_one("div#component_namemyfield_name")
    assert hidden_inputs_container.attrs["hx-swap-oob"] == "true"
    hidden_inputs = hidden_inputs_container.select(
        "input[type='hidden'][name='myfield_name']"
    )
    assert len(hidden_inputs) == 1
    assert hidden_inputs[0].attrs["value"] == str(to_add.id)

    # 3. The autocomplete input
    # this component should have mostly been tested in the items test
    text_input = soup.select_one("input#component_namemyfield_name__textinput")
    assert text_input.attrs["hx-vals"]
    assert "multiselect" not in text_input.attrs["hx-vals"]

    assert text_input.attrs["value"] == to_add.name

    # 4. The "data" span, not sure how this is used
    data_span = soup.select_one("span#component_namemyfield_name__data")
    assert data_span.attrs["data-phac-aspc-autocomplete"] == str(to_add.name)

    # 5. some script tag to keep events updated,
    script_tag = soup.select_one(
        "script[data-componentid='component_namemyfield_name']"
    )
    assert "phac_aspc_autocomplete_trigger_change" in script_tag.get_text()


def test_toggle_response_unselect_non_multi(client):
    """
    try toggling off a person that was selected
    """
    people = PersonFactory.create_batch(5)
    to_remove = PersonFactory()

    base_url = reverse("autocomplete:toggle", kwargs={"ac_name": "PersonAC3"})

    # next, try removign a person from a list with one person
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "myfield_name": to_remove.id,
            "component_prefix": "component_name",
            "item": to_remove.id,
        }
    )

    response = client.get(f"{base_url}?{qs_dict.urlencode()}")
    assert response.status_code == 200

    soup = get_soup(response)


def test_toggle_multi(client):
    people = PersonFactory.create_batch(5)
    p1 = PersonFactory()
    p2 = PersonFactory()

    base_url = reverse("autocomplete:toggle", kwargs={"ac_name": "PersonAC3"})
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "item": p1.id,
            "multiselect": True,
        }
    )

    response = client.get(f"{base_url}?{qs_dict.urlencode()}")
    assert response.status_code == 200

    soup = get_soup(response)

    # 1. The element we are toggling
    toggled_option = soup.select("a[role='option']")
    assert len(toggled_option) == 1

    # 2. hidden inputs
    hidden_inputs_container = soup.select_one("div#component_namemyfield_name")
    assert hidden_inputs_container.attrs["hx-swap-oob"] == "true"
    hidden_inputs = hidden_inputs_container.select(
        "input[type='hidden'][name='myfield_name']"
    )

    # 3. The autocomplete input
    assert soup.select_one("ul#component_namemyfield_name_ac_container")
    chips = soup.select("ul#component_namemyfield_name_ac_container > li.chip")
    chip_buttons = soup.select(
        "ul#component_namemyfield_name_ac_container > li.chip > a"
    )
    for chip_a in chip_buttons:
        assert "multiselect" in chip_a.attrs["hx-params"]
        assert "multiselect" in chip_a.attrs["hx-vals"]

    input_li = soup.select(
        "ul#component_namemyfield_name_ac_container > li.input:not(.chip)"
    )
    assert len(chips) == 1
    assert len(input_li) == 1

    # 4. The "info", I think this is a11y stuff
    info_text = soup.select_one("div#component_namemyfield_name__info").get_text()
    assert p1.name in info_text

    # 5. screen reader description, kinda redundant with 4
    sr_description = soup.select_one(
        "div#component_namemyfield_name__sr_description"
    ).get_text()
    assert p1.name in sr_description

    assert len(hidden_inputs) == 1
    values = {int(i.attrs["value"]) for i in hidden_inputs}
    assert values == {p1.id}

    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "multiselect": True,
            "myfield_name": p1.id,
            "item": p2.id,
        }
    )
    response = client.get(f"{base_url}?{qs_dict.urlencode()}")
    assert response.status_code == 200

    soup = get_soup(response)

    # hidden inputs:
    hidden_inputs = soup.select(
        "div#component_namemyfield_name input[type='hidden'][name='myfield_name']"
    )
    assert len(hidden_inputs) == 2
    values = {int(i.attrs["value"]) for i in hidden_inputs}
    assert values == {p1.id, p2.id}


def test_toggle_multi_untoggle(client):
    people = PersonFactory.create_batch(5)
    p1 = PersonFactory()
    p2 = PersonFactory()
    p3 = PersonFactory()

    base_url = reverse("autocomplete:toggle", kwargs={"ac_name": "PersonAC3"})

    # untoggle from a list of 3
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "item": p1.id,
            "multiselect": True,
        }
    )
    qs_dict.setlist("myfield_name", [p1.id, p2.id, p3.id])

    response = client.get(f"{base_url}?{qs_dict.urlencode()}")
    assert response.status_code == 200

    soup = get_soup(response)

    # 1. The element we are toggling
    toggled_option = soup.select("a[role='option']")
    assert len(toggled_option) == 1

    # 2. hidden inputs
    hidden_inputs = soup.select(
        "div#component_namemyfield_name input[type='hidden'][name='myfield_name']"
    )
    assert len(hidden_inputs) == 2
    values = {int(i.attrs["value"]) for i in hidden_inputs}
    assert values == {p2.id, p3.id}

    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "item": p1.id,
            "multiselect": True,
            "myfield_name": p1.id,
        }
    )
    resp = client.get(f"{base_url}?{qs_dict.urlencode()}")
    assert resp.status_code == 200
    soup = get_soup(resp)

    hidden_inputs = soup.select(
        "div#component_namemyfield_name input[type='hidden'][name='myfield_name']"
    )
    assert len(hidden_inputs) == 0



import json

from django.http import HttpRequest, QueryDict
from django.urls import reverse

from autocomplete.core import Autocomplete, register
from sample_app.models import Person, PersonFactory, Team, TeamFactory
from tests.conftest import PersonAC

from .utils_for_test import get_soup


def test_items_response_non_multi(client):

    people = PersonFactory.create_batch(5)
    searchable_person = PersonFactory(name="abcdefg")
    searchable_person2 = PersonFactory(name="abcdxyz")

    base_url = reverse("autocomplete:items", kwargs={"ac_name": "PersonAC"})
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "search": "abcd",
        }
    )
    full_url = f"{base_url}?{qs_dict.urlencode()}"

    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)

    listbox = soup.select_one("div[role='listbox']")
    assert listbox.attrs["id"] == "component_namemyfield_name__items"
    assert "aria-multiselectable" not in listbox.attrs

    # abcd should match two people,
    options = listbox.select("a")
    assert len(options) == 2
    assert "abcdefg" in options[0].get_text()
    assert "abcdxyz" in options[1].get_text()

    assert json.loads(options[0].attrs["hx-vals"]) == {
        "field_name": "myfield_name",
        "component_prefix": "component_name",
        "item": str(searchable_person.id),
    }
    assert json.loads(options[1].attrs["hx-vals"]) == {
        "field_name": "myfield_name",
        "component_prefix": "component_name",
        "item": str(searchable_person2.id),
    }

    highlight_span = listbox.select_one("span.highlight")
    assert highlight_span.get_text() == "abcd"

    assert options[0].attrs["hx-get"] == reverse(
        "autocomplete:toggle", kwargs={"ac_name": "PersonAC"}
    )
    assert (
        options[0].attrs["hx-params"] == "myfield_name,field_name,item,component_prefix"
    )
    assert options[0].attrs["hx-include"] == "#component_namemyfield_name"
    assert "component_namemyfield_name__item__" in options[0].attrs["id"]
    assert not options[1].attrs["id"] == options[0].attrs["id"]


def test_items_response_multi(client):

    class PersonAC2(Autocomplete):
        # registry is not cleared between tests, must use unique names

        @classmethod
        def search_items(cls, search, context):
            qs = Person.objects.filter(name__icontains=search)

            return [{"key": person.id, "label": person.name} for person in qs]

        @classmethod
        def get_items_from_keys(cls, keys, context):
            return Person.objects.filter(id__in=keys)

    register(PersonAC2)

    people = PersonFactory.create_batch(5)
    searchable_person = PersonFactory(name="abcdefg")
    searchable_person2 = PersonFactory(name="abcdxyz")

    base_url = reverse("autocomplete:items", kwargs={"ac_name": "PersonAC2"})
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "search": "abcd",
            "multiselect": True,
        }
    )
    full_url = f"{base_url}?{qs_dict.urlencode()}"

    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)

    listbox = soup.select_one("div[role='listbox']")
    assert listbox.attrs["id"] == "component_namemyfield_name__items"
    assert listbox.attrs["aria-multiselectable"] == "true"
    assert listbox.attrs["aria-description"] == "multiselect"

    # abcd should match two people,
    options = listbox.select("a")
    assert len(options) == 2
    assert "abcdefg" in options[0].get_text()
    assert "abcdxyz" in options[1].get_text()

    assert json.loads(options[0].attrs["hx-vals"]) == {
        "field_name": "myfield_name",
        "component_prefix": "component_name",
        "item": str(searchable_person.id),
        "multiselect": True,
    }
    assert json.loads(options[1].attrs["hx-vals"]) == {
        "field_name": "myfield_name",
        "component_prefix": "component_name",
        "item": str(searchable_person2.id),
        "multiselect": True,
    }

    highlight_span = listbox.select_one("span.highlight")
    assert highlight_span.get_text() == "abcd"

    assert options[0].attrs["hx-get"] == reverse(
        "autocomplete:toggle", kwargs={"ac_name": "PersonAC2"}
    )
    assert (
        options[0].attrs["hx-params"]
        == "myfield_name,field_name,item,component_prefix,multiselect"
    )
    assert options[0].attrs["hx-include"] == "#component_namemyfield_name"
    assert "component_namemyfield_name__item__" in options[0].attrs["id"]
    assert not options[1].attrs["id"] == options[0].attrs["id"]

    # now add abcdefg as member and try again,
    qs_dict.setlist("myfield_name", [searchable_person.id])
    full_url = f"{base_url}?{qs_dict.urlencode()}"

    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)

    options = soup.select("div[role='listbox'] > a")
    assert len(options) == 2
    assert "abcdefg" in options[0].get_text()
    assert "abcdxyz" in options[1].get_text()

    assert options[0].attrs["aria-selected"] == "true"


def test_custom_options(client):

    people = PersonFactory.create_batch(5)
    searchable_person = PersonFactory(name="abcdefg")
    searchable_person2 = PersonFactory(name="abcdxyz")

    base_url = reverse("autocomplete:items", kwargs={"ac_name": "PersonAC"})

    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "search": "abcd",
            "placeholder": "my placeholder",
            "required": True,
            "disabled": True,
        }
    )
    full_url = f"{base_url}?{qs_dict.urlencode()}"

    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)

    result = soup.select_one("div[role='listbox'] a[role='option']")
    hx_params = result.attrs["hx-params"].split(",")
    assert "required" in hx_params
    assert "disabled" in hx_params
    assert "placeholder" in hx_params

    hx_vals = json.loads(result.attrs["hx-vals"])
    assert hx_vals["required"]
    assert hx_vals["disabled"]
    assert hx_vals["placeholder"] == "my placeholder"


def test_limit_results(client):
    people = PersonFactory.create_batch(5)
    searchable_person = PersonFactory(name="abcdefg")
    searchable_person2 = PersonFactory(name="abcdxyz")
    searchable_person3 = PersonFactory(name="abcdxyz2")

    @register
    class LimitedPersonAC(PersonAC):
        max_results = 2
        narrow_search_text = "NARROW IT DOWN"

    base_url = reverse("autocomplete:items", kwargs={"ac_name": "LimitedPersonAC"})
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield",
            "search": "abcd",
        }
    )
    full_url = f"{base_url}?{qs_dict.urlencode()}"

    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)
    listbox = soup.select_one("div[role='listbox']")
    results = listbox.select("a")
    assert len(results) == 2

    more_results = listbox.select_one("div.more-results")
    assert "NARROW IT DOWN" in more_results.get_text()


def test_no_results(client):
    @register
    class NoResultsPersonAC(PersonAC):
        max_results = 2
        no_result_text = "NO RESULTS"

    base_url = reverse("autocomplete:items", kwargs={"ac_name": "NoResultsPersonAC"})
    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield",
            "search": "abcd",
        }
    )
    full_url = f"{base_url}?{qs_dict.urlencode()}"
    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)
    listbox = soup.select_one("div[role='listbox']")
    results = listbox.select("a")
    assert len(results) == 0
    items = listbox.select("span.item")
    assert len(items) == 1
    assert "NO RESULTS" in items[0].get_text()


def test_query_too_short(client):
    base_url = reverse("autocomplete:items", kwargs={"ac_name": "PersonAC"})

    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "search": "s",
            "placeholder": "my placeholder",
            "required": True,
            "disabled": True,
        }
    )
    full_url = f"{base_url}?{qs_dict.urlencode()}"

    response = client.get(full_url)
    assert response.status_code == 200

    soup = get_soup(response)

    listbox = soup.select_one("div[role='listbox']")
    results = listbox.select("a")
    assert len(results) == 0
    items = listbox.select("span.item")
    assert len(items) == 1
    assert "Type at least 3 characters" in items[0].get_text()



"""
Django template tags to facilitate rendering of the component
"""

import hashlib
import json

from django import template, urls
from django.template import loader
from django.template.defaultfilters import stringfilter
from django.utils.html import escape, format_html
from django.utils.http import urlencode
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
@stringfilter
def make_id(value):
    """Generate an ID given a string, to use as element IDs in HTML"""
    return hashlib.sha1(value.encode("utf-8")).hexdigest()


@register.filter()
@stringfilter
def search_highlight(value, search):
    """Surround the section of text matching the search with a classed span"""
    if search == "":
        return value
    try:
        pos = value.lower().index(search.lower())
        start = escape(value[:pos])
        match = escape(value[pos : pos + len(search)])
        end = escape(value[pos + len(search) :])
        return mark_safe(f'{start}<span class="highlight">{match}</span>{end}')
    except ValueError:
        pass
    return value


@register.simple_tag(takes_context=True)
def use_string(context, name, strings):
    """
    Loads the string from a template or via the variable dict `strings` if the `name`
    key is defined within.  This allows strings to be overriden in 2 ways, either by
    user defined templates which will override *all* instances, or via the
    `custom_strings` property of the Autocomplete instance which allows individual
    customization.

    When `name` is not found in `strings`, the template name becomes:

    autocomplete/strings/{name}.html

    """
    if name in strings:
        return strings[name]

    return loader.get_template(
        f"autocomplete/strings/{name}.html", using="django"
    ).render(context.flatten())


@register.simple_tag
def substitute_string(template_str, **kwargs):
    """
    Substitute the template string with the kwargs
    """
    as_strings = {k: str(v) for k, v in kwargs.items()}
    return template_str % as_strings


@register.simple_tag
def autocomplete(name, selected=None):
    """
    Tag used to render autocomplete component in a Django Template

    Parameters:

        name        Name of the component.
                    The name is used to generate routes and must be unique.
                    Required.

        selected    Comma separated list of selected values
                    Defaults to None

    """
    options_selected = (
        ",".join([str(x) for x in selected]) if selected is not None else ""
    )

    url = urls.reverse(name, kwargs={"method": "component"})
    parameter = urlencode({name: options_selected})
    get_url = f"{url}?{parameter}"

    return format_html(
        (
            "<div "
            f'hx-get="{get_url}"'
            'hx-trigger="load"'
            'hx-swap="outerHTML">'
            "</div>"
        )
    )


@register.filter
def js_boolean(value):
    """
    Convert the value to a javascript boolean
    """
    return "true" if value else "false"


@register.simple_tag
def autocomplete_head(bootstrap=False):
    """
    Renders the styles required for the component

    Parameters:

        bootstrap   Set to true if the bootstrap css should be loaded from cdn
                    Defaults to False.
    """
    return loader.get_template("autocomplete/head.html", using="django").render(
        {"bootstrap": bootstrap}
    )


@register.simple_tag(takes_context=True)
def autocomplete_scripts(context, bootstrap=False, htmx=False, htmx_csrf=False):
    """
    Renders the required script tags for the component

    Parameters:

        bootstrap   Set to true if bootstrap should be loaded from cdn
                    Defaults to False.

        htmx        Set to true if htmx should be loaded from cdn
                    Defaults to False.

        htmx_csrf   Set to true if htmx should be initialized with csrf token
                    Defaults to False.

    """
    return loader.get_template("autocomplete/scripts.html", using="django").render(
        {
            "csrf_token": context.get("csrf_token", ""),
            "bootstrap": bootstrap,
            "htmx": htmx,
            "htmx_csrf": htmx_csrf,
        }
    )


@register.simple_tag
def value_if_truthy(test, value, default=""):
    """
    Return the value if it is truthy, otherwise return the default
    """
    return value if test else default


@register.simple_tag(takes_context=True)
def base_configurable_values_hx_params(context):

    field_name = context.get("field_name")
    required = context.get("required")
    disabled = context.get("disabled")
    placeholder = context.get("placeholder")
    multiselect = context.get("multiselect")

    hx_params = f"{field_name},field_name,item,component_prefix"

    if required:
        hx_params += ",required"

    if disabled:
        hx_params += ",disabled"

    if placeholder:
        hx_params += ",placeholder"

    if multiselect:
        hx_params += ",multiselect"

    return mark_safe(hx_params)


@register.simple_tag(takes_context=True)
def base_configurable_hx_vals(context):
    """
    json-like format
    must be wrapped in curly braces
    """

    field_name = context.get("field_name")
    required = context.get("required")
    disabled = context.get("disabled")
    placeholder = context.get("placeholder")
    multiselect = context.get("multiselect")
    component_prefix = context.get("component_prefix")

    props = {
        "field_name": escape(field_name),
        "component_prefix": component_prefix,
    }

    if required:
        props["required"] = bool(required)

    if disabled:
        props["disabled"] = bool(disabled)

    if multiselect:
        props["multiselect"] = bool(multiselect)

    if placeholder:
        props["placeholder"] = escape(placeholder)

    hx_vals = json.dumps(props).replace("{", "").replace("}", "")

    return mark_safe(hx_vals)


def stringify_extra_hx_vals(extra_hx_vals_dict):
    if any("'" in val for val in extra_hx_vals_dict.values()):
        raise ValueError(
            "Extra hx vals cannot contain single quotes, consider backticks for JS expressions or escaping double-quotes"
        )

    return ",".join([f' "{key}": {val}' for key, val in extra_hx_vals_dict.items()])


@register.simple_tag(takes_context=True)
def text_input_hx_vals(context):
    """
    items has augments hx-vals,
    - it adds JS value of the search input
    - users can add more values in their class
    """

    base_hx_vals_str = base_configurable_hx_vals(context)

    component_id_escape = escape(context.get("component_id"))

    val = (
        "js:{"
        f"{base_hx_vals_str},"
        f'search: document.getElementById("{component_id_escape}__textinput").value'
    )

    extra_hx_vals = context.get("ac_class").get_extra_text_input_hx_vals()
    if extra_hx_vals:
        extra_hx_val_str = stringify_extra_hx_vals(extra_hx_vals)
        val = f"{val}, {extra_hx_val_str}"

    val = val + "}"

    return mark_safe(val)



"""app URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/4.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_sample_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from django.contrib import admin
from django.urls import include, path

from autocomplete import urls as autocomplete_urls
from sample_app import views

urlpatterns = [
    path("admin/", admin.site.urls),
    path("teams/<int:team_id>/edit/", views.edit_team, name="edit_team"),
    path(
        "teams/<int:team_id>/edit/with_prefix/",
        views.example_with_prefix,
        name="edit_team_w_prefix",
    ),
    path("teams/<int:team_id>/edit3/", views.example_with_model, name="edit_team3"),
    path("ac/", autocomplete_urls),
    path("app/__debug__/", include("debug_toolbar.urls")),
]



import pytest
from django import forms
from django.template import Context, Template, loader
from django.urls import reverse

from autocomplete import Autocomplete, AutocompleteWidget, register
from sample_app.models import Person, PersonFactory, Team, TeamFactory

from .utils_for_test import soup_from_str


@register
class PersonAC4(Autocomplete):

    @classmethod
    def search_items(cls, search, context):
        qs = Person.objects.filter(name__icontains=search)

        return [{"key": person.id, "label": person.name} for person in qs]

    @classmethod
    def get_items_from_keys(cls, keys, context):
        qs = Person.objects.filter(id__in=keys)
        return [{"key": person.id, "label": person.name} for person in qs]


single_form_template = Template(
    """
        {{ form.as_p }}
    """
)


def render_template(template, ctx_dict):
    context = Context(ctx_dict)
    return template.render(context)


def test_render_widget_in_form_empty():
    class FormWithSingle(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["team_lead"]

            widgets = {
                "team_lead": AutocompleteWidget(
                    ac_class=PersonAC4,
                )
            }

    create_form = FormWithSingle()

    rendered = render_template(single_form_template, {"form": create_form})

    soup = soup_from_str(rendered)

    # check the form label works
    label = soup.select_one("label[for='id_team_lead']")
    assert label.text == "Team lead:"

    # check focus ring,
    focus_ring = soup.select_one("div.phac_aspc_form_autocomplete_focus_ring")
    assert focus_ring

    component_container = focus_ring.select_one(
        "div.phac_aspc_form_autocomplete#team_lead__container"
    )
    assert component_container

    # 1. hidden input are in #<component_id>
    # it starts out empty without even a name
    component = component_container.select_one("#team_lead")
    inputs = component.select('span > input[type="hidden"]')
    assert len(inputs) == 1
    assert inputs[0].attrs["name"] == "team_lead"
    assert "value" not in inputs[0].attrs

    # 2. script
    scripts = soup.select("script")
    assert len(scripts) == 1
    assert scripts[0].attrs["data-componentid"] == "team_lead"
    assert scripts[0].attrs["data-toggleurl"] == reverse(
        "autocomplete:toggle", args=["PersonAC4"]
    )

    ac_container_ul = soup.select_one("ul#team_lead_ac_container.ac_container")
    assert ac_container_ul

    lis = ac_container_ul.select("li")
    assert len(lis) == 1
    actual_input_field = lis[0].select_one("input")
    assert actual_input_field.attrs["type"] == "text"
    assert "name" not in actual_input_field.attrs
    assert actual_input_field.attrs["aria-controls"] == "team_lead__items"
    assert actual_input_field.attrs["hx-get"] == reverse(
        "autocomplete:items", args=["PersonAC4"]
    )
    assert actual_input_field.attrs["hx-include"] == "#team_lead"
    assert actual_input_field.attrs["hx-target"] == "#team_lead__items"
    assert (
        'getElementById("team_lead__textinput")' in actual_input_field.attrs["hx-vals"]
    )
    assert '"component_prefix": "",' in actual_input_field.attrs["hx-vals"]
    assert '"field_name": "team_lead",' in actual_input_field.attrs["hx-vals"]
    assert "value" not in actual_input_field.attrs


def test_render_widget_in_form_non_empty():
    class FormWithSingle(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["team_lead"]

            widgets = {
                "team_lead": AutocompleteWidget(
                    ac_class=PersonAC4,
                )
            }

    lead = PersonFactory()
    record = TeamFactory(team_lead=lead)

    create_form = FormWithSingle(
        instance=record,
    )

    rendered = render_template(single_form_template, {"form": create_form})

    soup = soup_from_str(rendered)

    # check input is populated,
    input = soup.select_one("#team_lead  span > input[type='hidden']")
    assert input.attrs["name"] == "team_lead"
    assert input.attrs["value"] == str(lead.id)

    ac_container_ul = soup.select_one("ul#team_lead_ac_container.ac_container")
    assert ac_container_ul

    lis = ac_container_ul.select("li")
    assert len(lis) == 1
    actual_input_field = lis[0].select_one("input")
    assert actual_input_field.attrs["type"] == "text"
    assert actual_input_field.attrs["value"] == lead.name
    assert "multiselect" not in actual_input_field.attrs["hx-vals"]


def test_render_widget_multi_empty():
    class FormWithMulti(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["members"]

            widgets = {
                "members": AutocompleteWidget(
                    ac_class=PersonAC4, options={"multiselect": True}
                )
            }

    people = PersonFactory.create_batch(5)

    create_form = FormWithMulti()

    rendered = render_template(single_form_template, {"form": create_form})

    soup = soup_from_str(rendered)


def test_render_widget_multi_non_empty():

    class FormWithMulti(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["members"]

            widgets = {
                "members": AutocompleteWidget(
                    ac_class=PersonAC4, options={"multiselect": True}
                )
            }

    people = PersonFactory.create_batch(5)
    record = TeamFactory()
    record.members.set(people[:2])

    edit_form = FormWithMulti(instance=record)

    rendered = render_template(single_form_template, {"form": edit_form})

    soup = soup_from_str(rendered)

    component = soup.select_one("#members")
    hidden_inputs = component.select('span > input[type="hidden"]')

    assert len(hidden_inputs) == 2
    assert hidden_inputs[0].attrs["name"] == "members"
    values = {int(i.attrs["value"]) for i in hidden_inputs}
    assert values == {people[0].id, people[1].id}

    # check untoggle buttons,
    ac_items = soup.select("ul#members_ac_container.ac_container > li.chip")
    assert len(ac_items) == 2
    for li in ac_items:
        assert li.select_one("span").text in {people[0].name, people[1].name}
        delete_button = li.select_one("a")
        assert delete_button.attrs["hx-get"] == reverse(
            "autocomplete:toggle", args=["PersonAC4"]
        )
        assert (
            delete_button.attrs["hx-params"]
            == "members,field_name,item,component_prefix,required,multiselect,remove"
        )
        assert "multiselect" in delete_button.attrs["hx-vals"]
        assert delete_button.attrs["hx-vals"]
        assert delete_button.attrs["hx-include"] == "#members"
        assert delete_button.attrs["hx-swap"] == "delete"
        assert delete_button.attrs["hx-target"] == "this"
        assert delete_button.attrs["href"] == "#"


def test_with_formset():
    p1 = PersonFactory()
    p2 = PersonFactory()
    p3 = PersonFactory()
    people = [p1, p2, p3]
    team1 = TeamFactory(team_lead=p1)
    team2 = TeamFactory(team_lead=p2)
    team3 = TeamFactory(team_lead=None)

    class FormWithSingle(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["team_lead"]

            widgets = {
                "team_lead": AutocompleteWidget(
                    ac_class=PersonAC4,
                )
            }

    forms.modelformset_factory(Team, form=FormWithSingle)

    template = Template(
        """
        <div id='empty-form-container'>
        {{ formset.empty_form }}
        </div>
        <div id='management-form'>
        {{ formset.management_form }}
        </div>
        {% for form in formset %}
            <div class="form-container">
            {{ form.as_p }}
            </div>
        {% endfor %}
    """
    )

    formset = forms.modelformset_factory(Team, form=FormWithSingle, extra=0)(
        queryset=Team.objects.all()
    )

    rendered = render_template(template, {"formset": formset})

    soup = soup_from_str(rendered)

    empty_form = soup.select_one("#empty-form-container")
    assert empty_form

    formset_forms = soup.select(".form-container")
    for ix, form in enumerate(formset_forms):
        hidden_input = form.select_one(
            f"input[type='hidden'][name='form-{ix}-team_lead']"
        )
        if ix in (0, 1):
            assert hidden_input.attrs["value"] == str(people[ix].id)
        if ix == 2:
            assert not hidden_input


def test_custom_options():
    class FormWithSingle(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["team_lead"]

            widgets = {
                "team_lead": AutocompleteWidget(
                    ac_class=PersonAC4,
                    attrs={
                        "required": True,
                        "disabled": True,
                    },
                    options={
                        "component_prefix": "my_prefix",
                        "placeholder": "my placeholder",
                    },
                )
            }

    create_form = FormWithSingle()

    rendered = render_template(single_form_template, {"form": create_form})

    soup = soup_from_str(rendered)

    input = soup.select_one("ul li input[type='text']")

    assert input.attrs["placeholder"] == "my placeholder"
    assert "required" in input.attrs
    assert "disabled" in input.attrs

    hx_vals = input.attrs["hx-vals"]
    assert '"placeholder": "my placeholder"' in hx_vals
    assert '"required": true' in hx_vals
    assert '"disabled": true' in hx_vals


def test_custom_options_not_required():
    class FormWithSingle(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["team_lead"]

            widgets = {
                "team_lead": AutocompleteWidget(
                    ac_class=PersonAC4,
                )
            }

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields["team_lead"].required = False

    create_form = FormWithSingle()

    rendered = render_template(single_form_template, {"form": create_form})

    soup = soup_from_str(rendered)

    input = soup.select_one("ul li input[type='text']")

    assert "required" not in input.attrs

    hx_vals = input.attrs["hx-vals"]
    assert "required" not in hx_vals


def test_disabled_multi():

    class FormWithMulti(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["members"]

            widgets = {
                "members": AutocompleteWidget(
                    ac_class=PersonAC4, options={"multiselect": True}
                )
            }

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields["members"].required = False
            self.fields["members"].disabled = True

    people = PersonFactory.create_batch(5)

    record = TeamFactory()
    record.members.set(people[:2])

    edit_form = FormWithMulti()
    edit_form = FormWithMulti(instance=record)

    rendered = render_template(single_form_template, {"form": edit_form})

    soup = soup_from_str(rendered)

    ac_container = soup.select_one(".ac_container")
    chips = ac_container.select("li.chip")
    assert len(chips) == 2
    for chip in chips:
        assert not chip.select_one("a")
    input_li = ac_container.select_one("li.input")
    assert not input_li.select_one("input")
    assert input_li.select_one("output#members__textinput")


def test_extra_hx_vals():
    @register
    class PersonAC4WithHxVals(PersonAC4):
        @classmethod
        def get_extra_text_input_hx_vals(cls):
            return {"extra": '"value"'}

    class FormWithSingle(forms.ModelForm):
        class Meta:
            model = Team
            fields = ["team_lead"]

            widgets = {
                "team_lead": AutocompleteWidget(
                    ac_class=PersonAC4WithHxVals,
                )
            }

    create_form = FormWithSingle()

    rendered = render_template(single_form_template, {"form": create_form})

    soup = soup_from_str(rendered)

    input = soup.select_one("ul li input[type='text']")
    hx_vals = input.attrs["hx-vals"]

    assert '"extra": "value"' in hx_vals



# Generated by Django 4.1.4 on 2024-09-20 01:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=60)),
            ],
        ),
        migrations.CreateModel(
            name="Team",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=60)),
                (
                    "members",
                    models.ManyToManyField(
                        related_name="teams", to="sample_app.person"
                    ),
                ),
                (
                    "team_lead",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="lead_teams",
                        to="sample_app.person",
                    ),
                ),
            ],
        ),
    ]



from bs4 import BeautifulSoup
from django.http import QueryDict
from django.test.client import MULTIPART_CONTENT, Client


def get_soup(response):
    content = response.content
    return soup_from_str(content)


def soup_from_str(content):
    soup = BeautifulSoup(content, "html.parser")
    return soup


def put_request_as_querystring(client, url, data):
    """
    for some reason,
    htmx sends PUT requests as querystings??
    """
    qs = QueryDict(mutable=True)
    for k, v in data.items():
        if isinstance(v, list):
            qs.setlist(k, v)
        else:
            qs.update({k: v})

    query_string = qs.urlencode()
    response = client.put(url, data=query_string)
    return response



from dataclasses import dataclass

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

# This is the registry of registered autocomplete classes,
# i.e. the ones who respond to requests
_ac_registry = {}


AC_CLASS_CONFIGURABLE_VALUES = {
    "disabled",
    "no_result_text",
    "narrow_search_text",
    "minimum_search_length",
    "max_results",
    "component_prefix",
    "placeholder",
    "indicator",
}


def register(ac_class: type, route_name: str = None):
    if not route_name:
        route_name = ac_class.__name__

    ac_class.validate()

    if route_name in _ac_registry:
        raise ValueError(f"Autocomplete with name '{name}' is already registered.")

    ac_class.route_name = route_name

    _ac_registry[route_name] = ac_class

    return ac_class


class Autocomplete:

    no_result_text = _("No results found.")
    narrow_search_text = _(
        "Showing %(page_size)s of %(total)s items. Narrow your search for more results."
    )
    type_at_least_n_characters = _("Type at least %(n)s characters")
    minimum_search_length = 3
    max_results = 100
    component_prefix = ""

    @classmethod
    def auth_check(cls, request):
        """
        override to inspect request.user or whatever
        raise a PermissionDenied or SuspiciousOperation exception if needed
        """
        if (
            getattr(settings, "AUTOCOMPLETE_BLOCK_UNAUTHENTICATED", False)
            and not request.user.is_authenticated
        ):
            raise PermissionDenied("Must be logged in to use autocomplete")

        pass

    @classmethod
    def validate(cls):
        if not hasattr(cls, "search_items"):
            raise ValueError("You must implement a search_items method.")

        if not hasattr(cls, "get_items_from_keys"):
            raise ValueError("You must implement a get_items_from_keys method.")

    @classmethod
    def map_search_results(cls, items_iterable, selected_keys=None):
        """
        This must return a list of dictionaries with the keys "key", "label", and "selected"

        By default, we already expect search_items to return iterable of the form [{"key": "value", "label": "label"}]

        You can override this to consume paginable querysets or whatever
        """

        return [
            {  # this is the default mapping
                "key": str(i["key"]),
                "label": i["label"],
                "selected": i["key"] in selected_keys or str(i["key"]) in selected_keys,
            }
            for i in items_iterable
        ]

    @classmethod
    def get_custom_strings(cls):
        return {
            "no_results": cls.no_result_text,
            "more_results": cls.narrow_search_text,
            "type_at_least_n_characters": cls.type_at_least_n_characters,
        }

    @classmethod
    def get_extra_text_input_hx_vals(cls):
        """
        returns a dict of key/vals to go in the hx-vals attribute of the text input
        - must not contain single quotes
        - to support inline JS expressions, values are not quoted
        """

        return {}


@dataclass
class ContextArg:
    request: HttpRequest
    client_kwargs: dict



from setuptools import setup
from setuptools.command.install_lib import install_lib as _install_lib
from setuptools.command.sdist import sdist as _sdist
from distutils.command.build import build as _build
from distutils.cmd import Command


class compile_translations(Command):
    description = 'compile message catalogs to MO files via django compilemessages'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import os
        import sys
        from django.core.management.commands.compilemessages import Command
        c = Command()
        curdir = os.getcwd()
        os.chdir(os.path.realpath('autocomplete'))
        c.handle(locale=[], exclude=[], ignore_patterns=[], fuzzy=False, verbosity=0)
        os.chdir(curdir)


class build(_build):
    sub_commands = [('compile_translations', None)] + _build.sub_commands


class install_lib(_install_lib):
    def run(self):
        self.run_command('compile_translations')
        _install_lib.run(self)

class sdist(_sdist):
    def run(self):
        self.run_command('compile_translations')
        _sdist.run(self)


setup(
    cmdclass={
        'build': build,
        'install_lib': install_lib,
        'sdist': sdist,
        'compile_translations': compile_translations
    }
)


from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.urls import path
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views import View

from .core import AC_CLASS_CONFIGURABLE_VALUES, ContextArg, _ac_registry


class AutocompleteBaseView(View):
    @cached_property
    def ac_class(self):
        ac_name = self.kwargs["ac_name"]

        try:
            return _ac_registry[ac_name]

        except KeyError as e:
            raise ValueError(f"No registered autocomplete with name {ac_name}") from e

    def dispatch(self, request, *args, **kwargs):
        self.ac_class.auth_check(request)

        return super().dispatch(request, *args, **kwargs)

    @cached_property
    def request_dict(self):
        # convert the request's QueryDict into a regular dict
        return self.request.GET.dict()

    def get_field_name(self):
        return self.request_dict["field_name"]

    def get_component_id(self):
        prefix = self.get_configurable_value("component_prefix")

        return prefix + self.get_field_name()

    def get_configurable_value(self, key):
        if key in self.request_dict:
            return self.request.GET.get(key)

        if key in AC_CLASS_CONFIGURABLE_VALUES and hasattr(self.ac_class, key):
            return getattr(self.ac_class, key)

        return None

    def get_template_context(self):
        # many things will come from the request
        # others will be picked up from the AC class

        return {
            "route_name": self.ac_class.route_name,
            "ac_class": self.ac_class,
            "field_name": self.get_field_name(),
            "component_id": self.get_component_id(),
            "required": bool(self.get_configurable_value("required")),
            "placeholder": self.get_configurable_value("placeholder"),
            "indicator": self.get_configurable_value("indicator"),
            "custom_strings": self.ac_class.get_custom_strings(),
            "multiselect": bool(self.get_configurable_value("multiselect")),
            "component_prefix": self.get_configurable_value("component_prefix"),
            "disabled": bool(self.get_configurable_value("disabled")),
        }


class ToggleView(AutocompleteBaseView):
    def get(self, request, *args, **kwargs):
        field_name = self.request_dict["field_name"]

        current_items = self.request.GET.getlist(field_name)
        if current_items == ["undefined"] or current_items == [""]:
            current_items = []

        key_to_toggle = request.GET.get("item")

        if key_to_toggle is None:
            return HttpResponseBadRequest()

        new_selected_keys = list(current_items)

        is_multi = self.get_configurable_value("multiselect")

        if is_multi:
            if key_to_toggle in current_items:
                new_selected_keys.remove(key_to_toggle)
            else:
                new_selected_keys.append(key_to_toggle)
        else:
            if new_selected_keys == []:
                new_selected_keys = [key_to_toggle]
            else:
                new_selected_keys = []

        keys_to_fetch = set(new_selected_keys).union({key_to_toggle})

        context_obj = ContextArg(request=request, client_kwargs=request.GET)
        all_values = self.ac_class.get_items_from_keys(keys_to_fetch, context_obj)

        items = self.ac_class.map_search_results(all_values, new_selected_keys)

        # OOB is used if the user clicks the X on a chip,
        # to update the selected style of the option
        # if it is currently in the dropdown list
        swap_oob = request.GET.get("remove", False)

        target_item = next((x for x in items if x["key"] == key_to_toggle), None)

        new_items = [x for x in items if x["key"] in new_selected_keys]

        def sort_items(item):
            try:
                return current_items.index(f"{item['key']}")
            except ValueError:
                return len(new_items)

        new_items = sorted(new_items, key=sort_items)

        if target_item is None:
            raise ValueError("Requested item to toggle not found.")

        return render(
            request,
            "autocomplete/item.html",
            {
                **self.get_template_context(),
                "search": "",
                "values": new_selected_keys,
                "item_as_list": [target_item],
                "item": target_item,
                "toggle": new_items,
                "swap_oob": swap_oob,
            },
        )


class ItemsView(AutocompleteBaseView):
    def get(self, request, *args, **kwargs):
        context_obj = ContextArg(request=request, client_kwargs=request.GET)

        search_query = request.GET.get("search", "")
        search_results = self.ac_class.search_items(
            # or whatever
            search_query,
            context_obj,
        )

        field_name = self.get_configurable_value("field_name")
        selected_keys = request.GET.getlist(field_name)

        query_too_short = len(search_query) < self.ac_class.minimum_search_length

        if query_too_short:
            total_results = 0
            search_results = []

        else:
            total_results = len(search_results)
            if total_results > self.ac_class.max_results:
                search_results = search_results[: self.ac_class.max_results]

        items = self.ac_class.map_search_results(search_results, selected_keys)

        # render items ...
        return render(
            request,
            "autocomplete/item_list.html",
            {
                # note: name -> field_name
                **self.get_template_context(),
                "show": not (query_too_short),
                "query_too_short": query_too_short,
                "search": search_query,
                "items": items,
                "total_results": total_results,
                "minimum_search_length": self.ac_class.minimum_search_length,
            },
        )


urls = (
    [
        path("autocomplete/<str:ac_name>/items", ItemsView.as_view(), name="items"),
        path("autocomplete/<str:ac_name>/toggle", ToggleView.as_view(), name="toggle"),
    ],
    "autocomplete",
    "autocomplete",
)



from django.contrib.auth.models import User
from django.http import QueryDict
from django.test import override_settings
from django.urls import reverse

from sample_app.models import PersonFactory
from tests.conftest import PersonAC


def urls():
    p = PersonFactory()

    toggle_url = reverse(
        "autocomplete:toggle",
        kwargs={
            "ac_name": "PersonAC",
        },
    )

    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "item": p.id,
        },
    )
    toggle_url = f"{toggle_url}?{qs_dict.urlencode()}"

    items_url = reverse(
        "autocomplete:items",
        kwargs={
            "ac_name": "PersonAC",
        },
    )

    qs_dict = QueryDict(mutable=True)
    qs_dict.update(
        {
            "field_name": "myfield_name",
            "component_prefix": "component_name",
            "search": "abcd",
        },
    )
    items_url = f"{items_url}?{qs_dict.urlencode()}"

    return toggle_url, items_url


def test_auth_check_blocks_unauthenticated(client):

    with override_settings(AUTOCOMPLETE_BLOCK_UNAUTHENTICATED=True):
        response = client.get(urls()[0])
        assert response.status_code == 403
        response = client.get(urls()[1])
        assert response.status_code == 403


def test_disabled_auth_check_allows_unauthenticated(client):
    with override_settings(AUTOCOMPLETE_BLOCK_UNAUTHENTICATED=False):
        response = client.get(urls()[0])
        assert response.status_code == 200
        response = client.get(urls()[1])
        assert response.status_code == 200


def test_auth_check_enabled_allows_authenticated(client):
    u = User.objects.create(username="a")
    client.force_login(u)

    with override_settings(AUTOCOMPLETE_BLOCK_UNAUTHENTICATED=True):
        response = client.get(urls()[0])
        assert response.status_code == 200
        response = client.get(urls()[1])
        assert response.status_code == 200



import random

from django.db import transaction

from sample_app.models import Person, PersonFactory, Team, TeamFactory


@transaction.atomic
def run():
    Team.objects.all().delete()
    Person.objects.all().delete()

    people = PersonFactory.create_batch(200)
    teams = TeamFactory.create_batch(40)

    for team in teams:
        team_members = random.sample(people, random.randint(0, 8))
        if team_members:
            team.members.set(team_members)
            team.team_lead = random.choice(team_members)
            team.save()
